/*
 This file is part of GNU Taler
 (C) 2020 Taler Systems S.A.

 GNU Taler is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
 Foundation; either version 3, or (at your option) any later version.

 GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
 WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with
 GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */

/**
 * Imports.
 */
import {
  MerchantApiClient,
  PreparePayResultType,
  j2s,
} from "@gnu-taler/taler-util";
import { WalletApiOperation } from "@gnu-taler/taler-wallet-core";
import { CoinConfig } from "../harness/denomStructures.js";
import { GlobalTestState } from "../harness/harness.js";
import {
  createSimpleTestkudosEnvironmentV2,
  createWalletDaemonWithClient,
  makeTestPaymentV2,
  withdrawViaBankV2,
} from "../harness/helpers.js";

const coinCommon = {
  cipher: "RSA" as const,
  durationLegal: "3 years",
  durationSpend: "2 years",
  durationWithdraw: "7 days",
  feeDeposit: "TESTKUDOS:0",
  feeRefresh: "TESTKUDOS:0",
  feeRefund: "TESTKUDOS:0",
  feeWithdraw: "TESTKUDOS:0",
  rsaKeySize: 1024,
};

const coinConfigList: CoinConfig[] = [
  {
    ...coinCommon,
    name: "n1",
    value: "TESTKUDOS:1",
  },
  {
    ...coinCommon,
    name: "n5",
    value: "TESTKUDOS:5",
  },
];

/**
 * Run test for paying a merchant with balance locked behind a pending refresh.
 */
export async function runWalletBlockedPayMerchantTest(t: GlobalTestState) {
  // Set up test environment

  const { bank, exchange, merchant } = await createSimpleTestkudosEnvironmentV2(
    t,
    coinConfigList,
  );

  // Withdraw digital cash into the wallet.

  const { walletClient: w1 } = await createWalletDaemonWithClient(t, {
    name: "w1",
    persistent: true,
    config: {
      testing: {
        devModeActive: true,
      },
    },
  });

  await withdrawViaBankV2(t, {
    walletClient: w1,
    bank,
    exchange,
    amount: "TESTKUDOS:20",
  });

  await w1.call(WalletApiOperation.TestingWaitTransactionsFinal, {});

  // Prevent the wallet from doing refreshes by injecting a 5xx
  // status for all refresh requests.
  await w1.call(WalletApiOperation.ApplyDevExperiment, {
    devExperimentUri: "taler://dev-experiment/start-block-refresh",
  });

  // Do a payment that causes a refresh.
  await makeTestPaymentV2(t, {
    merchant,
    walletClient: w1,
    order: {
      summary: "test",
      amount: "TESTKUDOS:2",
    },
  });

  const merchantClient = new MerchantApiClient(merchant.makeInstanceBaseUrl());

  const orderResp = await merchantClient.createOrder({
    order: {
      summary: "My Payment",
      amount: "TESTKUDOS:18",
    },
  });

  let orderStatus = await merchantClient.queryPrivateOrderStatus({
    orderId: orderResp.order_id,
  });

  t.assertTrue(orderStatus.order_status === "unpaid");

  // Make wallet pay for the order

  const preparePayResult = await w1.call(WalletApiOperation.PreparePayForUri, {
    talerPayUri: orderStatus.taler_pay_uri,
  });

  console.log(`prepare pay result: ${j2s(preparePayResult)}`);

  t.assertTrue(
    preparePayResult.status === PreparePayResultType.PaymentPossible,
  );

  await w1.call(WalletApiOperation.ApplyDevExperiment, {
    devExperimentUri: "taler://dev-experiment/stop-block-refresh",
  });

  await w1.call(WalletApiOperation.TestingWaitTransactionsFinal, {});
}

runWalletBlockedPayMerchantTest.suites = ["wallet"];
